<?php

// PHP functions to generate an image tooltip that displays box coordinates.
//
// Copyright (C) 2006-2011 by Pawel Pilarczyk, http://www.pawelpilarczyk.com/
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU Affero
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public
// License along with this software; see the file "license.txt".
// If not, see <http://www.gnu.org/licenses/>.

// Started in 2006. Last revision: April 27, 2011.


// Returns HTML code that equips an image with a tooltip
// that shows the position of the cursor with respect to the given
// (possibly coarser) grid, and optionally also other values.
// Use negative values for 'boxwidth' and/or 'boxheight' to indicate
// how many boxes there are (e.g. -32 for a grid of 32 boxes).
// The border width is necessary in calculating the cursor position.
// Set 'values' to the name of an array defined in an earlier
// JavaScript block, or to 'false' if none should be used.
// If 'strict' is set to 1 then the tooltip is only enabled
// in browsers in which it is tested to display correct values,
// currently Mozilla Firefox only (sorry about that).
// Set 'debug' to 1 for additional info at tooltip, 0 for none.
// Example usage: <img $returnedCode src="image.png"/>
function imgtooltiphtml ($boxwidth, $boxheight, $border, $values, $formname,
	$strict, $debug)
{
	if ($values === false)
		$values = 'false';
	return "onmouseover=\"tooltipShow(this," .
		"$boxwidth,$boxheight,$border,$values,'$formname'," .
		"$strict,$debug);\"\nonmouseout=\"tooltipHide();\"";
} /* imgtooltiphtml */

// Returns a JavaScript block with tooltip code,
// to be put in the "head" section of a HTML page.
function imgtooltipcode ()
{
	return <<< _SCRIPT_

<!--
This javascript code uses parts of the public-domain code
and information provided at the following websites:
http://lixlpixel.org/javascript-tooltips/
http://www.quirksmode.org/js/findpos.html
http://www.quirksmode.org/js/events_properties.html
http://www.javascriptkit.com/javatutors/navigator.shtml
-->

<script type="text/javascript">
<!--

// detect if the browser is supported in the "strict" mode
var browserSupported =
	(navigator.userAgent.indexOf ('Firefox') >= 0) ||
	(navigator.userAgent.indexOf ('Chrome') >= 0) ||
	(navigator.userAgent.indexOf ('Safari') >= 0) ||
	(navigator.appName === 'Microsoft Internet Explorer');

// a reference to the tooltip object
var tooltip = false;

// Computes the position of the object relative to the document page
// and stores the computed values in the tooltip object.
function findObjectPosition (obj)
{
	if (tooltip === false)
		return;
	var left = 0;
	var top = 0;
	if (obj.offsetParent) {
		do {
			left += obj.offsetLeft;
			top += obj.offsetTop;
		} while (obj = obj.offsetParent);
	}
	tooltip.imgposx = left;
	tooltip.imgposy = top;
	return;
}

// Creates a new tooltip element.
function createTooltip ()
{
	if (!document.createElement)
		return;

	// create the tooltip as a "div" element
	tooltip = document.createElement ('div');
	tooltip.id = 'tooltip';

	// set the style of the tooltip
	tooltip.style.display = 'none';
	tooltip.style.position = 'absolute';
	tooltip.style.padding = '3px';
	tooltip.style.color = '#000000';
	tooltip.style.background = '#eeeeee';
	tooltip.style.border = '1px solid #aaaaaa';
	tooltip.style.textAlign = 'center';
	tooltip.style.fontSize = 'smaller';

	// set empty HTML code for the tooltip
	tooltip.innerHTML = '';

	// set the position of the tooltip relative to the mouse (in pixels)
	tooltip.offsetx = 14;
	tooltip.offsety = 10;

	// add the tooltip to the body elements
	document.body.appendChild (tooltip);
}

// Updates the tooltip and moves it to the mouse position.
var ie5 = (document.getElementById && document.all); 
var ns6 = (document.getElementById && !document.all); 
var ua = navigator.userAgent.toLowerCase();
var isapple = (ua.indexOf('applewebkit') != -1 ? 1 : 0);
function tooltipUpdate (e)
{
	if (tooltip === false)
		return;

	var documentBody = (document.compatMode &&
		document.compatMode != 'BackCompat') ? 
			document.documentElement : document.body;
	var pagex = ((isapple) ? 0 :
		((ie5) ? documentBody.scrollLeft : window.pageXOffset));
	var pagey = ((isapple) ? 0 :
		((ie5) ? documentBody.scrollTop : window.pageYOffset));
	var mousex = (ie5) ? event.x : ((ns6) ? clientX = e.clientX : false);
	var mousey = (ie5) ? event.y : ((ns6) ? clientY = e.clientY : false);

	var posx = mousex + pagex;
	var posy = mousey + pagey;
	if (!e)
		var e = window.event;

	// the following code block generates an error in MSIE, so in case
	// of that browser it must be desabled
	if (navigator.appName !== 'Microsoft Internet Explorer')
	{
		if (e.pageX || e.pageY)
		{
			posx = e.pageX;
			posy = e.pageY;
		}
		else if (e.clientX || e.clientY)
		{
			posx = e.clientX + document.body.scrollLeft
				+ document.documentElement.scrollLeft;
			posy = e.clientY + document.body.scrollTop
				+ document.documentElement.scrollTop;
		}
	}

	// set the position of the tooltip on the screen
	tooltip.style.left = (posx + tooltip.offsetx) + 'px';
	tooltip.style.top = (posy + tooltip.offsety) + 'px';

	// determine the position of mouse cursor within the image;
	// note: the coordinates are math-style, with the origin
	// in the bottom left corner
	var imgx0 = posx - tooltip.imgposx;
	var imgy0 = posy - tooltip.imgposy;

	// make a correction for the case of MSIE
	if (navigator.appName === 'Microsoft Internet Explorer')
	{
		imgx0 -= 1;
		imgy0 -= 1;
	}
	// make a correction for Mozilla Firefox
	if (navigator.userAgent.indexOf ('Firefox') >= 0)
	{
		if (!(tooltip.imgwidth % 2))
			imgx0 += 1;
	}

	var imgx = imgx0 - tooltip.border;
	var imgy = tooltip.imgheight - 1 - (imgy0 - tooltip.border);
	if ((imgx < -tooltip.border - 1) || (imgy < -tooltip.border - 1) ||
		(imgx > tooltip.imgwidth + tooltip.border + 1) ||
		(imgy > tooltip.imgheight + tooltip.border + 1))
	{
		tooltipHide ();
		return;
	}
	if (imgx < 0)
		imgx = 0;
	if (imgy < 0)
		imgy = 0;
	if (imgx >= tooltip.imgwidth - 1)
		imgx = tooltip.imgwidth - 1;
	if (imgy >= tooltip.imgheight - 1)
		imgy = tooltip.imgheight - 1;

	// calculate the coordinates to be displayed
	var boxx = Math.floor (imgx / tooltip.boxwidth);
	var boxy = Math.floor (imgy / tooltip.boxheight);

	// update the coordinates in the form if necessary
	if (tooltip.form)
	{
		tooltip.form.box_x.value = boxx;
		tooltip.form.box_y.value = boxy;
	}

	// prepare the additional value to be displayed after the coords
	var value = (tooltip.values === false) ? "" :
		(": " + tooltip.values [boxx] [boxy]);

	// prepare debug information if any
	var debug = '';
	if (tooltip.debug)
	{
		debug = "<br/>form xy = " + imgx0 + "," + imgy0 +
			"<br/>&mdash;&mdash;&mdash;&mdash;" +
			"<br/>calc xy = " + imgx + "," + imgy +
			"<br/>imgsize = " + tooltip.imgwidth + "x" +
			tooltip.imgheight +
			"<br/>border = " + tooltip.border +
			"<br/>boxsize = " + tooltip.boxwidth + "x" +
			tooltip.boxheight +
			"<br/>imgpos = " + tooltip.imgposx + "x" +
			tooltip.imgposy +
			"<br/>page = " + pagex + "," + pagey +
			"<br/>mouse = " + mousex + "," + mousey +
			"<br/>pos = " + posx + "," + posy;
	}

	// show the prepared data
	tooltip.innerHTML = "(" + boxx + "," + boxy + ")" + value + debug;

	return;
}

// Defines tooltip and turns it on.
// Arguments: the image object ('this'),
// image width and height (0 = autodetect),
// box width and height (negative: the number of boxes),
// 'values' - an array of values,
// 'debug' - should debugging info be displayed.
function tooltipShow (imgobj, boxwidth, boxheight, border, values, formname,
	strict, debug)
{
	// don't do anything in the strict mode
	// if the browser is not supported, unless debugging
	if (strict && !debug && !browserSupported)
		return;

	// create the tooltip object if it was not yet created
	if (tooltip === false)
		createTooltip ();

	// fill in the contents of the tooltip object
	tooltip.innerHTML = "&nbsp;";
	tooltip.style.display = 'block';
	tooltip.imgobj = imgobj;
	tooltip.imgwidth = imgobj.offsetWidth - 2 * border;
	tooltip.imgheight = imgobj.offsetHeight - 2 * border;
	if (boxwidth < 0)
		boxwidth = Math.ceil (tooltip.imgwidth / -boxwidth);
	if (boxheight < 0)
		boxheight = Math.ceil (tooltip.imgheight / -boxheight);
	tooltip.boxwidth = boxwidth;
	tooltip.boxheight = boxheight;
	tooltip.border = border;
	tooltip.values = values;
	if (formname && (formname != '') &&
		document.getElementById &&
		document.getElementById ('form_' + formname))
	{
		tooltip.form = document.getElementById ('form_' + formname);
	}
	else if (formname && (formname != '') &&
		document.forms['form_' + formname])
	{
		tooltip.form = document.forms['form_' + formname];
	}
	else
	{
		tooltip.form = 0;
	}
	tooltip.strict = strict;
	tooltip.debug = debug;
	findObjectPosition (imgobj);
//	tooltip.prevonmousemove = document.onmousemove;
	document.onmousemove = tooltipUpdate;
}

function tooltipHide ()
{
	if (tooltip === false)
		return;
	if (tooltip.debug)
		return;
	tooltip.style.display = 'none';
//	document.onmousemove = tooltip.prevonmousemove;
}

//-->
</script>
_SCRIPT_;
}

?>