#!/usr/bin/perl

#############################################################################
#
# This Perl script runs suitable programs and scripts in order to compute
# the index pairs and the index maps for all the previously computed
# combinatorial Morse sets.
# Please, adjust all the names hard-coded at the beginning of the script.
# Note that the specific internals of the map (if this is an ODE time-t map)
# must be already set in the corresponding file.
#
#############################################################################

# Copyright (C) 1997-2014 by Pawel Pilarczyk.
#
# This file is part of my research software package. This is free software:
# you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file "license.txt". If not,
# please, see <https://www.gnu.org/licenses/>.

# Started on March 9, 2013. Last revision: March 19, 2013.


# input files
my $logName = "_cm.log";
my $setPrefix = "_cm";
my $setSuffix = ".cub";

# output files with cubical sets
my $mainPrefix = "s";
my $mainSuffix = "q1.cub";
my $exitPrefix = "s";
my $exitSuffix = "q0.cub";
my $exitPlusPrefix = "s";
my $exitPlusSuffix = "q2.cub";

# output files with combinatorial maps
my $mapMainPrefix = "s";
my $mapMainSuffix = "m1.map";
my $mapExitPrefix = "s";
my $mapExitSuffix = "m0.map";
my $mapEntirePrefix = "s";
my $mapEntireSuffix = "m.map";

# logs
my $logPrefix = "";
my $logSuffix = ".log";
my $homPrefix = "";
my $homSuffix = ".hom";

# zipped files
my $zipMain = "_all.zip";
my $zipLogs = "_all_logs.zip";

# commands and programs
my $rename = "mv";
my $cat = "cat";
my $zip = "zip";

my $localProgPath = "./";
my $compmap = "${localProgPath}compmap";
#my $imgdisjoint = "${localProgPath}imgdisjoint";

my $chompProgPath = "~/chomp/bin/";
my $map2cub = "${chompProgPath}map2cub";
my $cubdiff = "${chompProgPath}cubdiff";
my $chkmvmap = "${chompProgPath}chkmvmap";
my $psetglue = "${chompProgPath}psetglue";
my $homcubes = "${chompProgPath}homcubes";

#####################################################

# runs the command
sub runCommand
{
	my $command = shift;
	print "$command\n";
	system ($command);
}

# determines the coordinates of the box, based on the log file of cmsingle
sub determineBox
{
	my $logFile;
	open ($logFile, $logName)
		or die "Can't open the log file '${logName}'.\n";
	my $firstLine = <$logFile>;
	$firstLine =~ m/cmsingle.* \-b ([\d,]+) /
		or die "Can't find the box coordinates in '${logName}'.\n";
	return $1;
}

# determines the file names of the computed combinatorial Morse sets
sub morseSets
{
	my @sets;
	foreach my $set (glob ("${setPrefix}*${setSuffix}"))
	{
		push (@sets, $1)
			if $set =~ m/${setPrefix}(\d+)${setSuffix}/;
	}
	return sort {$a <=> $b} @sets;
}

# computes the main map on the Morse set
sub computeMainMap
{
	my $set = shift;
	my $box = shift;
	my $cubName = "${mainPrefix}${set}${mainSuffix}";
	my $mapName = "${mapMainPrefix}${set}${mapMainSuffix}";
	my $logName = "${logPrefix}${mapName}${logSuffix}";
	my $cmd1 = "${compmap} -b ${box} " .
		"${cubName} ${mapName} " .
		"--debug --quiet --log ${logName}";
	runCommand ($cmd1);
}

# computes the exit set from the main map on the Morse set
sub computeExitSet
{
	my $set = shift;
	my $cubName = "${mainPrefix}${set}${mainSuffix}";
	my $mapName = "${mapMainPrefix}${set}${mapMainSuffix}";
	my $exitName = "${exitPrefix}${set}${exitSuffix}";

	my $logName1 = "${logPrefix}${exitName}${logSuffix}1";
	my $cmd1 = "${map2cub} " .
		"${mapName} -i ${exitName}.tmp " .
		"--quiet --log ${logName1}";
	runCommand ($cmd1);

	my $logName2 = "${logPrefix}${exitName}${logSuffix}2";
	my $cmd2 = "${cubdiff} " .
		"${exitName}.tmp ${cubName} ${exitName} " .
		"--quiet --log ${logName2}";
	runCommand ($cmd2);
}

# computes the map on the exit set
sub computeExitMap
{
	my $set = shift;
	my $box = shift;
	my $exitName = "${exitPrefix}${set}${exitSuffix}";
	my $cubName = "${mainPrefix}${set}${mainSuffix}";
	my $mapName = "${mapExitPrefix}${set}${mapExitSuffix}";
	my $logName = "${logPrefix}${mapName}${logSuffix}";

	my $cmd1 = "${compmap} -b ${box} " .
		"${exitName} ${mapName} -d ${cubName} " .
		"--debug --quiet --log ${logName}";
	runCommand ($cmd1);
}

# computes the joint map
sub computeJointMap
{
	my $set = shift;
	my $mapMainName = "${mapMainPrefix}${set}${mapMainSuffix}";
	my $mapExitName = "${mapExitPrefix}${set}${mapExitSuffix}";
	my $mapEntireName = "${mapEntirePrefix}${set}${mapEntireSuffix}";

	my $cmd1 = "${cat} ${mapMainName} ${mapExitName} > ${mapEntireName}";
	runCommand ($cmd1);
}

# computes the enlarged exit set
sub computeExitPlus
{
	my $set = shift;
	my $exitName = "${exitPrefix}${set}${exitSuffix}";
	my $mapExitName = "${mapExitPrefix}${set}${mapExitSuffix}";
	my $exitPlusName = "${exitPlusPrefix}${set}${exitPlusSuffix}";

	my $logName1 = "${logPrefix}${exitPlusName}${logSuffix}1";
	my $cmd1 = "${map2cub} " .
		"${mapExitName} -i ${exitPlusName}.tmp " .
		"--quiet --log ${logName1}";
	runCommand ($cmd1);

	my $logName2 = "${logPrefix}${exitPlusName}${logSuffix}2";
	my $cmd2 = "${psetglue} " .
		"${exitName} ${exitPlusName}.tmp ${exitPlusName} " .
		"--quiet --log ${logName2}";
	runCommand ($cmd2);
}

# checks if the maps are acyclic
sub checkAcyclic
{
	my $set = shift;
	my $cubName = "${mainPrefix}${set}${mainSuffix}";
	my $exitName = "${exitPrefix}${set}${exitSuffix}";
	my $exitPlusName = "${exitPlusPrefix}${set}${exitPlusSuffix}";
	my $mapName = "${mapEntirePrefix}${set}${mapEntireSuffix}";
	my $logName = "${logPrefix}${mapName}${logSuffix}";

	my $cmd1 = "${chkmvmap} -i " .
		"${mapName} ${cubName} ${exitName} " .
		"${cubName} ${exitPlusName} " .
		"--debug --quiet --log ${logName}";
	runCommand ($cmd1);
}

# computes the index map in homology
sub computeHomology
{
	my $set = shift;
	my $cubName = "${mainPrefix}${set}${mainSuffix}";
	my $exitName = "${exitPrefix}${set}${exitSuffix}";
	my $exitPlusName = "${exitPlusPrefix}${set}${exitPlusSuffix}";
	my $mapName = "${mapEntirePrefix}${set}${mapEntireSuffix}";
	my $homName = "${homPrefix}${mapName}${homSuffix}";

	my $cmd1 = "${homcubes} -i -a " .
		"${mapName} ${cubName} ${exitName} " .
		"${cubName} ${exitPlusName} " .
		"--debug --quiet --log ${homName}";
	runCommand ($cmd1);
}

# compresses all the main files
sub zipAllFiles
{
	my @sets = @_;

	my @files;
	foreach my $set (@sets)
	{
		my $cubName = "${mainPrefix}${set}${mainSuffix}";
		my $exitName = "${exitPrefix}${set}${exitSuffix}";
		my $mapMainName = "${mapMainPrefix}${set}${mapMainSuffix}";
		my $mapExitName = "${mapExitPrefix}${set}${mapExitSuffix}";
		push (@files, $cubName, $exitName,
			$mapMainName, $mapExitName);
	}

	my $cmd1 = "${zip} -9 ${zipMain} @files";
	runCommand ($cmd1);
}

# compresses all the log files
sub zipLogFiles
{
#	my @sets = @_;

	my @files = (glob ("*${logSuffix}"), glob ("*${logSuffix}1"),
		glob ("*${logSuffix}2"), glob ("*${logSuffix}3"),
		glob ("*${homSuffix}"), glob ("*${homSuffix}2"));
	my $files = join (" ", @files);

	my $cmd1 = "${zip} -9 -u ${zipLogs} ${files} -x _*.log";
	runCommand ($cmd1);
}

#####################################################

my $box = determineBox ();
my @sets = morseSets ();
print "=== Computing index pairs and cubical maps for ",
	scalar (@sets), " Morse sets for (${box}). ===\n";

foreach my $set (@sets)
{
	print "--- Processing Morse set no. ${set}. ---\n";

	my $cmd1 = "${rename} ${setPrefix}${set}${setSuffix} " .
		"${mainPrefix}${set}${mainSuffix}";
	runCommand ($cmd1) unless -e "${mainPrefix}${set}${mainSuffix}";

	computeMainMap ($set, $box);
	computeExitSet ($set);
	computeExitMap ($set, $box);
	computeJointMap ($set);
	computeExitPlus ($set);
}

print "=== Compressing the files. ===\n";

zipAllFiles (@sets);
zipLogFiles (@sets);

print "=== Checking acyclicity and computing homology for ",
	scalar (@sets), " Morse sets for (${box}). ===\n";

foreach my $set (@sets)
{
	print "--- Processing Morse set no. ${set}. ---\n";
	checkAcyclic ($set);
	computeHomology ($set);
}

print "=== Updating the logs archive. ===\n";

zipLogFiles (@sets);

print "=== Done. ===\n";

