#!/usr/bin/perl

#############################################################################
#
# This Perl script runs commands which create a nice continuation diagram.
# Call with two optional argumets: cube set prefix, coord.dat file name.
# Requires the following programs: convert, pset2bmp, psetjoin.
#
#############################################################################

# Copyright (C) 1997-2022 by Pawel Pilarczyk.
#
# This file is part of my research software package. This is free software:
# you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file "license.txt". If not,
# please, see <https://www.gnu.org/licenses/>.

# Started in 2007. Last revision: June 28, 2022.
# Output file names changed on February 20, 2022. Requires new "pset2bmp".


# the distance for joining (use 0 for no joining)
my $joinDistance = 40;

# the number of largest sets to skip (not to join with any other set)
my $joinSkipsets = 12;

# the size of boxes in pixels (spacing defined below is added)
my @boxSizes = (1, 6, 20);
my $boxSize = $boxSizes [0];

# the pause between boxes
my @pauses = (0, 1, 1);
my $pause = $pauses [0];

# determine the data directory and the port number
my $dirbase = `./workparam.pl dirbase`;
my $datdir = `./workparam.pl datdir`;
my $cmgdir = `./workparam.pl cmgdir`;
my $zipdir = `./workparam.pl zipdir`;
my $datname = `./workparam.pl name`;
my $datnumber = `./workparam.pl number`;
$datnumber = "";
print "dirbase = ${dirbase}, name = ${name}, number = ${number}\n";

# set the correct commands to use
my $psetjoin = "psetjoin";
$psetjoin = "./psetjoin" if -e "./psetjoin";
my $pset2bmp = "pset2bmp";
$pset2bmp = "./pset2bmp" if -e "./pset2bmp";

# the prefix of the names of sets of cubes
my $cubPrefix = "${datdir}/cont";
my $cubPath = $datdir;
if (scalar (@ARGV) > 0)
{
	$cubPrefix = $ARGV[0];

	# the path of the sets of cubes and other files
	$cubPrefix =~ m[(^.*)/];
	$cubPath = $1;
}

# determine the name of a file with the coordinator data
my $coordDat = "${cubPath}/coord.dat";
$coordDat = $ARGV[1] if scalar (@ARGV) > 1;

# determine the size of the background set of cubes
my $xRange = 0;
my $yRange = 0;
my $zRange = 0;
if (-f $coordDat)
{
	my $file;
	open ($file, "<$coordDat") or die "Can't open the file '$coordDat'";
	while (<$file>)
	{
		next unless m/^; subdivisions: (\d+) x (\d+)/;
		$xRange = $1;
		$yRange = $2;
		$zRange = $3 if m/^; subdivisions: (\d+) x (\d+) x (\d+)/;
		last;
	}
	close $file;
}
die "Can't determine the ranges of cubes"
	if (($xRange <= 0) || ($yRange <= 0));

# don't join in the 3-dimensional case
$joinDistance = 0 if $zRange;

# the prefix of joined sets of cubes
my $joinPrefix = $joinDistance ? ($cubPrefix . "j") : $cubPrefix;

# join the sets of cubes
if ($joinDistance)
{
	# delete the previous result of joining the sets
	foreach my $file (glob ("${joinPrefix}*"))
	{
		unlink $file if -f $file;
	}

	# join the sets
	my $cmd1 = "${psetjoin} ${cubPrefix}*.cub ${joinPrefix} " .
		"-d ${joinDistance}";
	$cmd1 .= " -s ${joinSkipsets}" if $joinSkipsets;
	print "$cmd1\n";
	print `$cmd1`;
}

# plot the 2-dimensional diagram
if (!$zRange)
{
	for (my $i = 0; $i < scalar (@boxSizes); ++ $i)
	{
		# the name of the output BMP and PNG files
		my $index = (scalar (@boxSizes) > 1) ? $i : '';
		my $bmpName = "${cubPrefix}diag${index}.bmp";
		my $pngName = "${cubPrefix}diag${index}.png";

		# choose the current box size and pause
		my $boxSize = $boxSizes [$i];
		my $pause = ($i < scalar (@pauses)) ?
			$pauses [$i] : $pauses [0];

		# plot the [joined] sets of cubes in a BMP file
		my $cmd2 = "${pset2bmp} ${joinPrefix}*.cub ${bmpName} " .
			"-s ${boxSize} -p ${pause} -r 300 -l 0 -b 0 " .
			"-w ${xRange} -h ${yRange} --pycolors";
		print "$cmd2\n";
		print `$cmd2`;

		# convert the output BMP file into a PNG file
		my $cmd3 = "convert ${bmpName} ${pngName}";
		# OLD: "nconvert -out png -o ${pngName} -overwrite ${bmpName}";
		print "$cmd3\n";
		print `$cmd3`;
	}
}

# the name of the output BMP and PNG files
my $bmpName = $cubPrefix . "diag.bmp";
my $pngName = $cubPrefix . "diag.png";

# plot slices of the 3-dimensional diagram
for (my $z = 0; $z < $zRange; ++ $z)
{
	# determine the file names for the slice pictures
	my $sliceNum = $z;
	$sliceNum = "0" . $sliceNum if (($z < 10) && ($zRange > 10));
	$sliceNum = "0" . $sliceNum if (($z < 100) && ($zRange > 100));
	my $bmpSlice = $bmpName;
	$bmpSlice =~ s/\.bmp$/${sliceNum}\.bmp/;
	my $pngSlice = $pngName;
	$pngSlice =~ s/\.png$/${sliceNum}\.png/;

	# if the PNG file already exists then skip it
	next if -f $pngSlice;

	# plot each slice in a BMP file
	my $cmd2 = "${pset2bmp} ${joinPrefix}*.cub ${bmpSlice} -s ${boxSize} " .
		"-p ${pause} -r 300 -l 0 -b 0 -w ${xRange} -h ${yRange} " .
		"--slice $z --pycolors";
	print "$cmd2\n";
	print `$cmd2`;

	# convert the output BMP file into a PNG file
	my $cmd3 = "convert ${bmpSlice} ${pngSlice}";
	# OLD: "nconvert -out png -o ${pngSlice} -overwrite ${bmpSlice}";
	print "$cmd3\n";
	print `$cmd3`;
}

# create the background set of cubes
my $backgroundName = "${cubPrefix}.bgr";
my $bgFile;
open ($bgFile, ">$backgroundName") or die "Can't create a background file";
for (my $x = 0; $x < $xRange; ++ $x)
{
	for (my $y = 0; $y < $yRange; ++ $y)
	{
		print $bgFile "(${x},${y})\n";
	}
}
close $bgFile;

# plot the 2-dimensional set of bad cubes
if (!$zRange || !(-e "${cubPrefix}.bad"))
{
	for (my $i = 0; $i < scalar (@boxSizes); ++ $i)
	{
		# set up the file names for the images
		# of the sets of bad cubes
		my $index = (scalar (@boxSizes) > 1) ? $i : '';
		my $badBmpName = "${cubPrefix}diag_bad${index}.bmp";
		my $badPngName = "${cubPrefix}diag_bad${index}.png";

		# choose the current box size and pause
		my $boxSize = $boxSizes [$i];
		my $pause = ($i < scalar (@pauses)) ?
			$pauses [$i] : $pauses [0];

		# skip bad cubes if there are none
		my $badCubName = "${cubPrefix}.bad";
		$badCubName = "skip" unless -e $badCubName;

		# plot the picture in a BMP file
		my $cmd4 = "${pset2bmp} skip skip ${badCubName} " .
			"skip " x 10 .
			"${backgroundName} ${badBmpName} -s ${boxSize} " .
			"-p ${pause} -r 300"; # note: previous colors
		print "$cmd4\n";
		print `$cmd4`;

		# convert the BMP file to a PNG one
		my $cmd5 = "convert ${badBmpName} ${badPngName}";
		#OLD: "nconvert -out png -o ${badPngName} -overwrite ${badBmpName}";
		print "$cmd5\n";
		print `$cmd5`;
	}
}

# plot slices of the 3-dimensional set of bad cubes
for (my $z = 0; $z < $zRange; ++ $z)
{
	# determine the file names for the slice pictures
	my $sliceNum = $z;
	$sliceNum = "0" . $sliceNum if (($z < 10) && ($zRange > 10));
	$sliceNum = "0" . $sliceNum if (($z < 100) && ($zRange > 100));
	my $bmpSlice = $badBmpName;
	$bmpSlice =~ s/\.bmp$/${sliceNum}\.bmp/;
	my $pngSlice = $badPngName;
	$pngSlice =~ s/\.png$/${sliceNum}\.png/;

	# if the PNG file already exists then skip it
	next if -f $pngSlice;

	# plot each slice in a BMP file
	my $cmd4 = "${pset2bmp} skip skip ${cubPrefix}.bad " . "skip " x 10 .
		"${backgroundName} ${bmpSlice} -s ${boxSize} -p ${pause} " .
		"-r 300 --slice $z"; # note: previous colors
	print "$cmd4\n";
	print `$cmd4`;

	# convert the BMP file to a PNG one
	my $cmd5 = "convert ${bmpSlice} ${pngSlice}";
	# OLD: "nconvert -out png -o ${pngSlice} ${bmpSlice}";
	print "$cmd5\n";
	print `$cmd5`;
}

# prepare the two PHP files for the browser in the 2D case
if (!$zRange)
{
	print "Preparing the two PHP files for the interactive browser...\n";
	my $output = <<'HTML';
<?php

// PHP web page for browsing the data acquired from 2-parameter computations.

require ("diagrams.php");

$name = "NNN";
$number = "MMM";
echo diagrams ("Continuation Diagram",
	"${name}${number}.php",
	array ("${name}${number}.png"),
	array (VVV, WWW), // the number of parameter intervals in each direction
	array (XXX, YYY), // the size of each square or rectangle in pixels
	array ("${name}${number}p.php?"),
	array ("[classes of equivalent Morse decompositions]"));

?>
HTML
	my $diagNumber = scalar (@boxSizes) - 1;
	`cp -p ${cubPrefix}diag${diagNumber}.png ${datdir}/${datname}${datnumber}.png`;
	$output =~ s/NNN/${datname}/;
	$output =~ s/MMM/${datnumber}/;
	$output =~ s/VVV/$xRange/;
	$output =~ s/WWW/$yRange/;
	my $xBoxSize = $boxSizes [scalar (@boxSizes) - 1] + $pauses [scalar (@pauses) - 1];
	my $yBoxSize = $xBoxSize;
	$output =~ s/XXX/$xBoxSize/;
	$output =~ s/YYY/$yBoxSize/;
	my $php1File;
	open ($php1File, ">${datdir}/${datname}${datnumber}.php")
		or die "Can't create the first PHP file";
	print $php1File $output;
	close $php1File;

	$output = <<'HTML';
<?php

// PHP web page for browsing the data acquired from the 2-parameter computations.

require ("browser.php");

$name = "NNN";
$number = "MMM";
echo browser (2, array (VVV, WWW), 0, array (
	"pngunzip.php?zip=zipped/${name}${number}cmg.zip&txt=g",
	"pngunzip.php?zip=zipped/${name}${number}png.zip&png=p"),
	array (".txt", ".png"),
	array ("Conley-Morse graph", "phase space portrait"),
	600);

?>
HTML
	$output =~ s/NNN/${datname}/;
	$output =~ s/MMM/${datnumber}/;
	$output =~ s/VVV/$xRange/;
	$output =~ s/WWW/$yRange/;
	my $php2File;
	open ($php2File, ">${datdir}/${datname}${datnumber}p.php")
		or die "Can't create the first PHP file";
	print $php2File $output;
	close $php2File;
}

