#!/usr/bin/perl

#############################################################################
#
# This script verifies every few seconds whether the processes
# named "cmgraphs" don't use too much memory. If this happens
# then it kills the processes which "blow up" in the memory.
# If no programs with the given name are found to be running
# then the script quits its job.
# The processes to be watched are determined within a few initial seconds:
# those that started in that period of time are taken into consideration,
# or all processes are taken if their list doesn't change within that time.
#
#############################################################################

# Copyright (C) 1997-2016 by Pawel Pilarczyk.
#
# This file is part of my research software package. This is free software:
# you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file "license.txt". If not,
# please, see <http://www.gnu.org/licenses/>.

# Started on February 24, 2008. Last revision: March 18, 2016.


# the name of a program to check for memory usage and to kill if necessary
my $prog = "cmgraphs";
$prog = $ARGV[0] if scalar (@ARGV) > 0;

# the maximal allowed memory usage of the program
my $maxmemAllowed = "2g";
$maxmemAllowed = $ARGV[1] if scalar (@ARGV) > 1;

# the time period (in minutes) for showing maximal memory usage
my $showstatevery = 1;

my $dir = `pwd`;
my $name = `echo \${USER}`;
$name =~ s/\s+$//;
my $name7 = ($name =~ m/(\S\S\S\S\S\S\S)\S\S/) ? $1 : $name;
die "No user name could be determined.\n" unless $name;

sub mem2mb
{
	my $memory = shift;
	if ($memory =~ m/([\d\.]+)\s*([mMgG])/)
	{
		my $memvalue = $1;
		my $memunits = $2;
	#	return int ($memvalue / 1024) if ($memunits =~ m/[kK]/);
		return int ($memvalue) if ($memunits =~ m/[mM]/);
		return int ($memvalue * 1024) if ($memunits =~ m/[gG]/);
		return 0;
	}
	return int ($memory / 1024);
}

my %firstprocesses;
my %processes;
my %warned;
my %maxused;
my %curused;
my $firsttime = 2;
my $showstatdelay = $showstatevery;
my $maxmem = mem2mb ($maxmemAllowed);
sub memguard
{
	my $count = 0;
	my @kill = ();
	my @list = `top -u ${name} -n 1 -b`;
	my @numbers;
	foreach my $line (@list)
	{
		next unless $line =~
			m/^\s*(\d+)\s*${name7}\S*\s*\d+\s*\d+\s*(\S+).*$prog/;
		my $number = $1;
		next if ((scalar keys %processes) &&
			(!defined $processes{$number}));
		my $memory = mem2mb ($2);
		push @numbers, $number;
		$curused{$number} = $memory;
		$maxused{$number} = $memory if ($maxused{$number} < $memory);
		if ($firsttime)
		{
			my $new = (($firsttime == 1) &&
				!defined ($firstprocesses{$number})) ?
				" [new!]" : "";
			print "Process $number uses ${memory} MB$new.\n";
			$firstprocesses{$number} = $memory
				if ($firsttime == 2);
		}
		elsif ($memory > $maxmem)
		{
			push @kill, $number;
			print "Killing $number: ${memory} MB.\n\n";
		}
		elsif (($memory > 0.9 * $maxmem) &&
			($memory != $warned{$number}))
		{
			print "Warning: $number uses ${memory} MB.\n";
			$warned{$number} = $memory;
		}
		++ $count;
	}

	if (!$firsttime)
	{
		if (-- $showstatdelay <= 0)
		{
			print "-----\n";
			for (my $i = 0; $i <= $#numbers; ++ $i)
			{
				my $number = $numbers[$i];
				my $curmemory = $curused{$number};
				my $maxmemory = $maxused{$number};
				print "Process ${number}: ${curmemory} MB " .
					"(max ${maxmemory} MB).\n";
			}
			$showstatdelay = $showstatevery + 1;
		}
	}

	if ($firsttime == 1)
	{
		# determine the new processes
		for (my $i = 0; $i <= $#numbers; ++ $i)
		{
			my $number = $numbers[$i];
			next if defined $firstprocesses{$number};
			$processes{$number} = 1 ;
		}

		# special version: if no new processes found then cancel
		die "No new processes found.\n"
			unless (scalar (keys %processes));

		# if no new processes were found, take all of them
		unless (scalar (keys %processes))
		{
			for (my $i = 0; $i <= $#numbers; ++ $i)
			{
				my $number = $numbers[$i];
				$processes{$number} = 1;
			}

			# increase the limit a little in order to avoid
			# killing a process that is guarded by another script
			$maxmem += int ($maxmem / 10);
		}

		# show which processes are being guarded against
		my $proclist = join ", ", sort {$a <=> $b} keys %processes;
		print "-----\n";
		print "Watching the following processes: $proclist.\n";
	}

	if ($firsttime)
	{
		print "-----\n";
		-- $firsttime;
	}

	foreach my $number (@kill)
	{
		`kill $number`;
	}

	# return the number of processes upon consideration
	return $count;
}

print "Guarding memory used by '${prog}' (max allowed: ${maxmem} MB).\n";
print "-----\n";
memguard ();
print "Waiting 1 sec...\n";
sleep 1;
print "-----\n";

my $noprocloops = 4;
while (1)
{
	my $count = memguard ();

	if (!$count)
	{
		-- $noprocloops;
		last if ($noprocloops <= 0);
	}
	else
	{
		$noprocloops = 4;
	}

	sleep 60;
}

print "Memory guarding finished.\n";

