/////////////////////////////////////////////////////////////////////////////
///
/// @file runmpiworkers.c
///
/// This is a wrapper for running workers using the MPI.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2022 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <https://www.gnu.org/licenses/>.

// Started on February 13, 2022. Last revision: February 14, 2022.

// Compile with: mpicc -O2 -o runmpiworkers runmpiworkers.c
// MPI part based on: https://usc-rc.github.io/tutorials/open-mpi
// Run with: mpirun ./runmpiworkers num0 pernode scriptprefix

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

const char *helpinfo = "\
This program calls a collection of workers using MPI.\n\
Call this program in the following way:\n\
  mpirun PROGRAM scriptprefix num0 pernode outprefix,\n\
where the arguments have the following meaning:\n\
num0 - the number of the first worker to launch,\n\
pernode - the number of workers per node,\n\
scriptprfix - the prefix of the script name for the workers.\n\
The program creates scripts for the workers by appending 4-digit\n\
0-padded number of each worker to the given prefix.\n\
It waits 2n seconds before launching the n-th worker at each node.\n\
The output of the script is redirected to a file with the given prefix.\n\
See the code and the comments therein for more information\n\
or ask the author; see www.pawelpilarczyk.com.\n\
";

int main(int argc, char** argv)
{
	// initialize the MPI environment
	MPI_Init(NULL, NULL);
	// determine the number of processes
	int world_size;
	MPI_Comm_size(MPI_COMM_WORLD, &world_size);
	// determine the rank of the current process
	int world_rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
	// get the name of the processor (the node running the process)
	char processor_name[MPI_MAX_PROCESSOR_NAME];
	int name_len;
	MPI_Get_processor_name(processor_name, &name_len);

	// check if the number of arguments is sufficient
	// and show help info if necessary
	if (argc < 5)
	{
		if (world_rank == 0)
			printf (helpinfo);
		exit (1);
	}

	// read the command-line attributes
	const char *progname = argv[0];
	const char *scriptprefix = argv[1];
	int num0 = atoi (argv[2]);
	int pernode = atoi (argv[3]);
	const char *outprefix = argv[4];

	// prepare the name of the script to run
	#define MAX_COMMAND 1000
	char command[MAX_COMMAND];
	if (strlen(scriptprefix)+strlen(outprefix)+40 > MAX_COMMAND)
	{
		printf ("ERROR: Too many characters for the command line "
			"(%d is the limit).\n", MAX_COMMAND);
		exit (-1);
	}
	int num = num0 + world_rank;
	if ((num < 0) || (num > 9999))
	{
		printf ("ERROR: Wrong worker's number %d.\n"
			"Should be beween 0 and 9999, inclusive).\n",
			num);
		exit (-1);
	}
	sprintf(command, "%s%04d < /dev/null > %s%04d.out 2>&1",
		scriptprefix, num, outprefix, num);

	// wait an appropriate number of seconds before running the script
	if (pernode <= 0)
	{
		printf ("ERROR: Wrong value (%d) of the number of processes "
			"per node.\n", pernode);
		exit (-1);
	}
	int seconds = 2 * (world_rank % pernode);
	if (seconds > 0)
		sleep (seconds);

	//printf("Hello world from processor %s, rank %d out of %d processors.\n",
	//	processor_name, world_rank, world_size);

	// run the script
	printf ("Running '%s' at %s after a delay of %d seconds (mpi process %d/%d).\n",
		command, processor_name, seconds, world_rank, world_size);
	system (command);

	// finish MPI environment
	MPI_Finalize();
	return (0);
} /* main */

