#!/usr/bin/perl

#############################################################################
#
# This script runs the coordinator using SLURM.
# It is based on the earlier script "runpbscoord.pl".
# See the script "run0pbs" for the details of the coordinator's command.
#
#############################################################################

# Copyright (C) 1997-2022 by Pawel Pilarczyk.
#
# This file is part of my research software package. This is free software:
# you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file "license.txt". If not,
# please, see <http://www.gnu.org/licenses/>.

# Started on April 13, 2011. Last revision: February 14, 2022.


# check the command line argument if the ssh connection should be used
# instead of submitting the jobs to the queue
my $ssh = ((scalar (@ARGV) > 0) && ($ARGV[0] eq 'ssh'));
# determine if the work should be done locally (as opposed to a SLURM job)
my $local = ((scalar (@ARGV) > 0) && ($ARGV[0] eq 'local'));
# determine if this only a test at the cluster
my $test = ((scalar (@ARGV) > 0) && ($ARGV[0] eq 'test'));

# determine various data for the computations
my $host = `./workparam.pl host`;
my $port = `./workparam.pl port`;
my $curdir = `./workparam.pl curdir`;
my $datdir = `./workparam.pl datdir`;
my $cubdir = `./workparam.pl cubdir`;
my $pngdir = `./workparam.pl pngdir`;
my $odedir = `./workparam.pl odedir`;
my $decdir = `./workparam.pl decdir`;
my $postprocdir = `./workparam.pl postprocdir`;
my $cmgdir = `./workparam.pl cmgdir`;
my $sharedir = `./workparam.pl sharedir`;
my $rundir = `./workparam.pl rundir`;
my $name = `./workparam.pl name`;
my $user = `./workparam.pl user`;
my $scheduler = `./workparam.pl scheduler`;
print "name = ${name}, coord = ${host}:${port}, data = ${datdir}\n";

# run the computations locally if there no scheduler could be determined
if ($scheduler eq "none")
{
	print "NOTE: Running the computations locally.";
	$local = 1;
}

# be careful if (some of) the directories already exist
if (-d $datdir)
{
	my $logfilename = "$datdir/coord.log";
	if (-f $logfilename)
	{
		print "Coordinator's log already exists: $logfilename\n";
		print "Rename it in order to avoid data loss.\n";
		exit;
	}
}

# create all the necessary directories
mkdir $datdir unless -d $datdir;
mkdir $pngdir unless -d $pngdir;
mkdir $odedir unless -d $odedir;
mkdir $decdir unless -d $decdir;
mkdir $cubdir unless -d $cubdir;
mkdir $postprocdir unless -d $postprocdir;
mkdir $cmgdir unless -d $cmgdir;
mkdir $sharedir unless -d $sharedir;
mkdir $rundir unless -d $rundir;

# copy all the files to the run dir
my $copycmd = "cp -p * ${rundir}";
#print "$copycmd\n";
system ($copycmd);

if ($ssh or $local)
{
	my $node = $host;
	unless (-e "node.txt")
	{
		my $nodefile;
		open ($nodefile, ">node.txt");
		print $nodefile $node;
		close $nodefile;
	}
	my $file;
	open ($file, "run0pbs")
		or die "Unable to open 'run0pbs'";
	my $script;
	read ($file, $script, 100000)
		or die "Unable to read 'run0pbs'";
	close $file;

	my $scriptname = "_run0ssh";
	open ($file, ">$scriptname")
		or die "Unable to create '$scriptname'";
	print $file <<"_SCRIPT_";
#!/bin/bash

DATDIR=${datdir}
CUBDIR=${cubdir}
PNGDIR=${pngdir}
ODEDIR=${odedir}
DECDIR=${decdir}
POSTPROCDIR=${postprocdir}
CMGDIR=${cmgdir}
SHAREDIR=${sharedir}
PORT=${port}
PBS_O_WORKDIR=${curdir}
PBS_O_HOST=${node}
PBS_O_QUEUE=none
PBS_QUEUE=none
PBS_ENVIRONMENT=none
PBS_JOBID=none
PBS_JOBNAME=none
PBS_NODEFILE=${curdir}/node.txt
PBS_O_HOME=${curdir}

_SCRIPT_
	print $file $script;
	close $file;
	chmod 0755, $scriptname;
	my $sshcommand =
		($local ? "sh ./" : "ssh ${node} ${curdir}/") .
		"${scriptname} > ${datdir}/coord.out 2>&1 " .
		"< /dev/null &";
#	print "$sshcommand\n";
	system ($sshcommand);
	print "The coordinator process started at ${node}.\n";
	exit;
}

# determine the walltime
my $walltime = "--time=72:00:00 ";
$walltime = "--time=00:15:00 " if $test;

# determine the right queue (partition) for the task
my $partition = "--partition=serial ";
$partition = "--partition=test " if $test;

# determine the memory amount that should be specified
my $memory = "--mem=4000M ";

# prepare a command for submitting the coordinator script to the PBS queue
my $command = "sbatch " .
	"--export=LD_LIBRARY_PATH=${ENV{LD_LIBRARY_PATH}}," .
	"DATDIR=${datdir},CUBDIR=${cubdir}," .
	"PNGDIR=${pngdir},ODEDIR=${odedir},DECDIR=${decdir}," .
	"POSTPROCDIR=${postprocdir}," .
	"CMGDIR=${cmgdir},SHAREDIR=${sharedir}," .
	"PORT=${port} " .
	"--output=${datdir}/coord.out " .
	"--ntasks=1 --mail-type=NONE " .
	$walltime . $memory . $partition .
	"--job-name=COR_${name} --no-requeue ./run0pbs";

# print and submit the coordinator script to the PBS queue
print "$command\n";
system ($command);

# wait until the coordinator has started
print "Waiting for the coordinator to start: ";
my $delay = 2;
my $seconds = 0;
my $pervflush = $|;
$| = 1;
while (1)
{
	sleep $delay;
	$seconds += $delay;
	my $jobid;
	my $state = "";
	my $host;
	my @jobs = `squeue --user=${user} --format="%.18i %.30j %.8u %.2t %.10M %.6D %R"`;
	foreach (@jobs)
	{
		# squeue sample line: 25164078     batch MW.30-.7   osmark  R    1:51:36      1 gv1336
		next unless m/^\s*(\d+).*COR\_${name}\s+${user}\s+(\S+).*\s(\S+)\s*$/;
		$jobid = $1;
		$state = $2;
		$host = $3;
		last;
	}
	if ($state eq "")
	{
		print "Not found.\n";
		last;
	}
	elsif ($state eq "R")
	{
		print " ($host) $seconds sec.\n";
		last;
	}
	else
	{
		print '.';
	}
}
print "\n";
$| = $prevflush;

