#!/usr/bin/perl

#############################################################################
#
# This script runs a set of workers to work for a coordinator which has
# already been launched (see "runslurmcoord.pl").
# The script is based on the previous script "runpbswork.pl".
# The workers are either submitted to a PBS queue or run locally
# (the 'local' argument).
# The number of workers are hard coded in the script;
# please, modify this data according to your needs.
# See the script "run1pbs" for the details of each worker's commands.
#
#############################################################################

# Copyright (C) 1997-2022 by Pawel Pilarczyk.
#
# This file is part of my research software package. This is free software:
# you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file "license.txt". If not,
# please, see <http://www.gnu.org/licenses/>.

# Started on February 23, 2008. Last revision: February 14, 2022.


# check the command line argument if the ssh connection should be used
# instead of submitting the jobs to the PBS queue
my $ssh = ((scalar (@ARGV) > 0) && ($ARGV[0] eq 'ssh'));
# determine if the work should be done locally (as opposed to a PBS job)
my $local = ((scalar (@ARGV) > 0) && ($ARGV[0] eq 'local'));
# determine if this only a test at the cluster
my $test = ((scalar (@ARGV) > 0) && ($ARGV[0] eq 'test'));
# determine if the MPI wrapper should be used
my $mpi = ((scalar (@ARGV) > 0) && ($ARGV[0] eq 'mpi'));

# determine the data directory and the port number
my $host = `./workparam.pl host`;
my $port = `./workparam.pl port`;
my $curdir = `./workparam.pl curdir`;
my $datdir = `./workparam.pl datdir`;
my $name = `./workparam.pl name`;
my $scheduler = `./workparam.pl scheduler`;
print "name = ${name}, coord = ${host}:${port}, data = ${datdir}\n";

# determine the command for the 'nice' program (with path)
my $nicecmd = "nice";
$nicecmd = "/bin/nice" if -x "/bin/nice";
$nicecmd = "/usr/bin/nice" if -x "/usr/bin/nice";

# run the computations locally if there no scheduler could be determined
if ($scheduler eq "none")
{
	print "NOTE: Running the computations locally.\n";
	$local = 1;
}

# switch to MPI if the host is "tryton-ap"
if (!$local and ($ENV{"HOSTNAME"} eq "tryton-ap"))
{
	print "NOTE: Using the MPI wrapper at Tryton.\n";
	$mpi = 1;
}

# determine if walltime should be specified
my $walltime = "";
$walltime = "--time=72:00:00 ";
#$walltime = "--time=16:00:00 ";
$walltime = "--time=00:15:00 " if $test;

# determine the right queue (partition) for the task
my $partition = "";
$partition = "--partition=batch ";
#$partition = "--partition=batch_16h ";
$partition = "--partition=test " if $test;

# determine if memory amount should be specified
my $memory = "";
$memory = "--mem-per-cpu=4500M ";

# determine the number of workers to submit
# NOTE: they all should fit in one node if using MPI
my $tasks = 120;
$tasks = 50 if $test;
$tasks = 8 if $local;

# determine the first available number for a worker to submit
# (assume that all the submitted workers are already running)
my $firstnumber = 0;
foreach (glob "${datdir}/work*")
{
	m/${datdir}\/work0*(\d+)\./ or next;
	my $num = $1;
	$firstnumber = $num + 1 if ($firstnumber <= $num);
}
my $lastnumber1 = $firstnumber + $tasks - 1;
print "Workers ${firstnumber}-${lastnumber1} will be submitted.\n";

# prepare local scripts for each worker if necessary
my @scriptnames;
if ($ssh or $local or $mpi)
{
	print "Creating local scripts for the $tasks workers...\n";

	# read the original script for running the worker
	my $file;
	open ($file, "run1pbs")
		or die "Unable to open 'run1pbs'";
	my $script;
	read ($file, $script, 100000)
		or die "Unable to read 'run1pbs'";
	close $file;

	# prepare scripts for all the workers
	for (my $n = $firstnumber; $n <= $lastnumber1; ++ $n)
	{
		# determine the 4-digit number of the worker
		my $num = $n;
		$num = '0' . $num if ($n < 1000);
		$num = '0' . $num if ($n < 100);
		$num = '0' . $num if ($n < 10);

		# create the script for this worker
		my $scriptname = "_run1ssh${num}";
		open ($file, ">$scriptname")
			or die "Unable to create '$scriptname'";

		# prepend the original script with some environment variables
		print $file <<"_SCRIPT_";
#!/bin/bash

NICECMD=${nicecmd}
DATDIR=${datdir}
NUM=${num}
COORD=${host}:${port}
PBS_O_WORKDIR=${curdir}
PBS_O_HOST=${node}
PBS_O_QUEUE=none
PBS_QUEUE=none
PBS_ENVIRONMENT=none
PBS_JOBID=none
PBS_JOBNAME=none
PBS_NODEFILE=${curdir}/node.txt
PBS_O_HOME=${curdir}

_SCRIPT_
		print $file $script;
		close $file;
		chmod 0755, $scriptname;

		push (@scriptnames, $scriptname);
	}
}

# run the local scripts for each worker if necessary
if ($ssh or $local)
{
	print "Launching the workers...\n";

	# create the node definition file if necessary
	$node = $host unless $node;
	unless (-e "node.txt")
	{
		my $nodefile;
		open ($nodefile, ">node.txt");
		print $nodefile $node;
		close $nodefile;
	}

	my $workerDelay = 2; # wait time between ssh or local workers are started
	my $submitted = 0;
	foreach my $scriptname (@scriptnames)
	{
		my $sshcommand =
			($local ? "sh ./" : "ssh ${node} ${curdir}/") .
			"${scriptname} > ${datdir}/work${num}.out 2>&1 " .
			"< /dev/null &";
	#	print "$sshcommand\n";
		system ($sshcommand);
		if ($workerDelay > 0)
		{
			$| ++;
			print (".");
			$| --;
			sleep ($workerDelay);
		}
		++ $submitted;
	}

	print "\n" if ($workerDelay > 0);
	print "${submitted} jobs started locally.\n";
	exit;
}

my $command = "";
if ($mpi)
{
	# prepare a command for submitting the MPI set of worker tasks to the SLURM queue
	$command = "sbatch --export=LD_LIBRARY_PATH=${ENV{LD_LIBRARY_PATH}}," .
		"NUM0=${firstnumber},DATDIR=${datdir},COORD=${host}:${port},NICECMD=${nicecmd} " .
		"--output=${datdir}/mpi${firstnumber}.out --mail-type=NONE " .
		"--ntasks=${tasks} --cpus-per-task=1 " .
		$walltime . $memory . $partition .
		"--job-name=CMG_${name} --no-requeue ./run1mpi";
}
else
{
	# prepare a command for submitting an array of worker scripts to the SLURM queue
	$command = "sbatch --array=${firstnumber}-${lastnumber1} " .
		"--export=LD_LIBRARY_PATH=${ENV{LD_LIBRARY_PATH}}," .
		"DATDIR=${datdir},COORD=${host}:${port},NICECMD=${nicecmd} " .
		"--output=${datdir}/work%A_%a.out --mail-type=NONE " .
		"--ntasks=1 --cpus-per-task=1 " .
		$walltime . $memory . $partition .
		"--job-name=CMG_${name} --no-requeue ./run1pbs";
}

# print and submit the worker scripts to the SLURM queue
print "$command\n";
system ($command);

